unit NumToText;
{<
  NumToText unit contains functions for converting of numeric value to text
  in Czech or English language.
  It is used by RTFReport and XMLReport component but can be used also independently.

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Sep 2006)
  @lastmod(Apr 2014)
}

interface

{
  Convert numeric value to English text
  Parameter Separator defines separator between words;
  Parameter Currency defines name of currency; '' = without currency, 'short' = dollars,  'long' = dollars
  Parameter SubCurrency defines name of SubCurrency; '' = cents
}
function NumToTextEN(Value: double; Separator: string = ''; Currency: string = ''; SubCurrency: string = ''): string;
{
  Convert numeric value to Czech text
  Parameter Separator defines separator between words;
  Parameter Currency defines name of currency; '' = without currency, 'short' = korun,  'long' =  korun ceskych
  Parameter SubCurrency defines name of SubCurrency; '' = haleru
}
function NumToTextCZ(Value: double; Separator: string = ''; Currency: string = ''; SubCurrency: string = ''): string;

implementation

uses
  SysUtils, Math;

type
  TNumToTextUniv = class
  public
    Value: double;
    Text: string;
    Separator: string;
    Currency: string;
    SubCurrency: string;
    procedure AddText(v: string);
    procedure AddCurrency; virtual; abstract;
    procedure AddSubCurrency; virtual; abstract;
    procedure AddNegative; virtual; abstract;
    procedure AddDigit(num: integer; exp: integer); virtual; abstract;
    procedure AddTens(num: integer; exp: integer); virtual; abstract;
    procedure AddHundreds(num: integer; exp: integer); virtual; abstract;
    procedure AddExp(num: integer; exp: integer); virtual; abstract;
    procedure AddZero; virtual; abstract;
    procedure Convert;
  end;

  TNumToTextEN = class(TNumToTextUniv)
    procedure AddCurrency; override;
    procedure AddSubCurrency; override;
    procedure AddNegative; override;
    procedure AddDigit(num: integer; exp: integer); override;
    procedure AddTens(num: integer; exp: integer); override;
    procedure AddHundreds(num: integer; exp: integer); override;
    procedure AddExp(num: integer; exp: integer); override;
    procedure AddZero; override;
  end;

  TNumToTextCZ = class(TNumToTextUniv)
    procedure AddCurrency; override;
    procedure AddSubCurrency; override;
    procedure AddNegative; override;
    procedure AddDigit(num: integer; exp: integer); override;
    procedure AddTens(num: integer; exp: integer); override;
    procedure AddHundreds(num: integer; exp: integer); override;
    procedure AddExp(num: integer; exp: integer); override;
    procedure AddZero; override;
  end;

function NumToTextLang(Lang: string; Value: double; Separator: string = ''; Currency: string = ''; SubCurrency: string = ''): string;
var
  obj: TNumToTextUniv;
begin
  if SameText(Lang, 'CZ') then
    obj:=TNumToTextCZ.Create
  else
    obj:=TNumToTextEN.Create;
  try
    obj.Value:=Value;
    obj.Separator:=Separator;
    obj.Currency:=Currency;
    obj.SubCurrency:=SubCurrency;
    obj.Convert;
    Result:=obj.Text;
  finally
    FreeAndNil(obj);
  end;
end;

function NumToTextEN(Value: double; Separator: string = ''; Currency: string = ''; SubCurrency: string = ''): string;
begin
  Result:=NumToTextLang('EN', Value, Separator, Currency, SubCurrency);
end;

function NumToTextCZ(Value: double; Separator: string = ''; Currency: string = ''; SubCurrency: string = ''): string;
begin
  Result:=NumToTextLang('CZ', Value, Separator, Currency, SubCurrency);
end;

// univ class
procedure TNumToTextUniv.AddText(v: string);
begin
  if v = '' then
    Exit;
  if Text <> '' then
    Text:=Text + Separator;
  Text:=Text + v;
end;

procedure TNumToTextUniv.Convert;
var
  e, vint: integer;
begin
  Text:='';
  if Value < 0 then
  begin
    AddNegative;
    Value:=-Value;
  end;
  if Value >= 1E15 then
    raise Exception.Create('Number out of range');

  if Value < 1 then
  begin
    AddZero;
  end
  else
    for e:=5 downto 1 do
    begin
      vint:=(Trunc(Value) mod Round(Power(10, e * 3))) div Round(Power(10, (e - 1) * 3));
      if vint > 0 then
      begin
        if vint >= 100 then
          AddHundreds(vint div 100, e);
        AddTens(vint mod 100, e);
        AddExp(vint, e);
      end;
    end;

  AddCurrency;

  vint:=Round(Value * 100) mod 100;
  if vint <> 0 then
  begin
    AddTens(vint, 0);
    AddSubCurrency;
  end;
end;

// EN class
procedure TNumToTextEN.AddNegative;
begin
  AddText('minus');
end;

procedure TNumToTextEN.AddDigit(num: integer; exp: integer);
begin
  case num of
    1:
      AddText('One');
    2:
      AddText('Two');
    3:
      AddText('Three');
    4:
      AddText('Four');
    5:
      AddText('Five');
    6:
      AddText('Six');
    7:
      AddText('Seven');
    8:
      AddText('Eight');
    9:
      AddText('Nine');
    10:
      AddText('Ten');
    11:
      AddText('Eleven');
    12:
      AddText('Twelve');
    13:
      AddText('Thirteen');
    14:
      AddText('Fourteen');
    15:
      AddText('Fifteen');
    16:
      AddText('Sixteen');
    17:
      AddText('Seventeen');
    18:
      AddText('Eighteen');
    19:
      AddText('Nineteen');
  end;
end;

procedure TNumToTextEN.AddTens(num: integer; exp: integer);
begin
  if num < 20 then
    AddDigit(num, exp)
  else
  begin
    case num div 10 of
      2:
        AddText('Twenty');
      3:
        AddText('Thirty');
      4:
        AddText('Forty');
      5:
        AddText('Fifty');
      6:
        AddText('Sixty');
      7:
        AddText('Seventy');
      8:
        AddText('Eighty');
      9:
        AddText('Ninety');
    end;
    AddDigit(num mod 10, 0);
  end;
end;

procedure TNumToTextEN.AddHundreds(num: integer; exp: integer);
begin
  AddDigit(num, 1);
  AddText('Hundred');
end;

procedure TNumToTextEN.AddExp(num: integer; exp: integer);
begin
  case exp of
    2:
      AddText('Thousand');
    3:
      AddText('Million');
    4:
      AddText('Billion');
    5:
      AddText('Trillion ');
  end;
end;

procedure TNumToTextEN.AddCurrency;
var
  vint: integer;
  sh, lng: boolean;
begin
  if Currency = '' then
    Exit;
  sh:=SameText(Currency, 'short');
  lng:=SameText(Currency, 'long');
  if sh or lng then
  begin
    vint:=Trunc(Value) mod 100;
    case vint of
      1:
        AddText('Dollar');
      else
        AddText('Dollars');
      end;
  end
  else
    AddText(Currency);
end;

procedure TNumToTextEN.AddSubCurrency;
begin
  if SubCurrency = '' then
    AddText('Cents')
  else
    AddText(SubCurrency);
end;

procedure TNumToTextEN.AddZero;
begin
  AddText('Zero');
end;

// CZ
procedure TNumToTextCZ.AddNegative;
begin
  AddText('mnus');
end;

procedure TNumToTextCZ.AddDigit(num: integer; exp: integer);
begin
  case num of
    1:
      case exp of
        0, 1, 4:
          AddText('jedna')
        else
          AddText('jeden');
      end;
    2:
      case exp of
        1, 4:
          AddText('dv')
        else
          AddText('dva');
      end;
    3:
      AddText('ti');
    4:
      AddText('tyi');
    5:
      AddText('pt');
    6:
      AddText('est');
    7:
      AddText('sedm');
    8:
      AddText('osm');
    9:
      AddText('devt');
    10:
      AddText('deset');
    11:
      AddText('jedenct');
    12:
      AddText('dvanct');
    13:
      AddText('tinct');
    14:
      AddText('trnct');
    15:
      AddText('patnct');
    16:
      AddText('estnct');
    17:
      AddText('sedmnct');
    18:
      AddText('osmnct');
    19:
      AddText('devatenct');
  end;
end;

procedure TNumToTextCZ.AddTens(num: integer; exp: integer);
begin
  if num < 20 then
    AddDigit(num, exp)
  else
  begin
    case num div 10 of
      2:
        AddText('dvacet');
      3:
        AddText('ticet');
      4:
        AddText('tyicet');
      5:
        AddText('padest');
      6:
        AddText('edest');
      7:
        AddText('sedmdest');
      8:
        AddText('osmdest');
      9:
        AddText('devadest');
    end;
    AddDigit(num mod 10, 0);
  end;
end;

procedure TNumToTextCZ.AddHundreds(num: integer; exp: integer);
begin
  if num > 1 then
    AddDigit(num, 1);
  case num of
    1:
      AddText('sto');
    2:
      AddText('st');
    3, 4:
      AddText('sta');
    5 .. 9:
      AddText('set');
  end;
end;

procedure TNumToTextCZ.AddExp(num: integer; exp: integer);
begin
  case exp of
    2:
      begin
        case num of
          2 .. 4:
            AddText('tisce');
          else
            AddText('tisc');
          end;
      end;
    3:
      begin
        case num of
          1:
            AddText('milin');
          2 .. 4:
            AddText('miliny');
          else
            AddText('milin');
          end;
      end;
    4:
      begin
        case num of
          1:
            AddText('miliarda');
          2 .. 4:
            AddText('miliardy');
          else
            AddText('miliard');
          end;
      end;
    5:
      begin
        case num of
          1:
            AddText('bilin');
          2 .. 4:
            AddText('biliny');
          else
            AddText('bilin');
          end;
      end;
  end;
end;

procedure TNumToTextCZ.AddCurrency;
var
  vint: integer;
  sh, lng: boolean;
begin
  if Currency = '' then
    Exit;
  sh:=SameText(Currency, 'short');
  lng:=SameText(Currency, 'long');
  if sh or lng then
  begin
    vint:=Trunc(Value) mod 100;
    case vint of
      1:
        AddText('koruna');
      2 .. 4:
        AddText('koruny');
      else
        AddText('korun');
      end;
    if lng then
      case vint of
        1:
          AddText('esk');
        2 .. 4:
          AddText('esk');
        else
          AddText('eskch');
        end;
  end
  else
    AddText(Currency);
end;

procedure TNumToTextCZ.AddSubCurrency;
begin
  if SubCurrency = '' then
    AddText('hal')
  else
    AddText(SubCurrency);
end;

procedure TNumToTextCZ.AddZero;
begin
  AddText('nula');
end;

end.
